"""Enumerations for Diematic."""
from enum import IntEnum


class DiematicStatus(IntEnum):
    """Represent the ENUMs of a status response."""

    OK = 0x0000
    ERROR = 0x0001


class DiematicOperation(IntEnum):
    """Represent the ENUMs of an operation."""

    GET_VALUES = 0x0001
    GET_CONFIG = 0x0002
    SET_VALUE  = 0x0003
    GET_VALUE  = 0x0004


"""Models for Diematic Boiler."""
from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass(frozen=True)
class Info:
    """Object holding information from Diematic"""

    model: str | None = None
    uuid: str | None = None
    manufacturer: str = "DeDietrich"
    name: str = "DeDietrich boiler"
    version: str = "0.0.0"

    @staticmethod
    def from_dict(data: dict):
        """Obtains Info from values in the data dictionary"""
        return Info(model="TBD", uuid=data["uuid"])


@dataclass(frozen=True)
class Boiler:
    """This is a Boiler converted into a dataclass, I'm not sure this is required at all"""

    info: Info
    variables: dict[str, Any]

    @staticmethod
    def from_dict(data):
        """Return a Boiler from a dictionary"""
        return Boiler(
            info=Info.from_dict(data), variables=Boiler.merge_variables_data(data)
        )

    @staticmethod
    def merge_variables_data(data: dict):
        """Obtain variables from the json"""
        variables = {}
        for k, value in data.items():
            if k == "uuid":
                continue
            variables[k] = value

        return variables



"""Asyncronous Python client for Diematic."""
from .boiler import (
	DiematicError,
	DiematicBoilerClient,
	DiematicConnectionError,
	DiematicParseError,
	DiematicResponseError,
	DiematicStatus,
	Boiler
)


"""Asyncronous Python client for DiematicServer"""
from unittest.mock import patch
import aiohttp
import async_timeout
import asyncio
from socket import gaierror as SocketGIAError
import logging

from .enums import (
    DiematicOperation,
    DiematicStatus
)

from .__version__ import __version__
from yarl import URL
from .models import Boiler
from deepmerge import always_merger
from typing import Any, Optional

from .exceptions import (
    DiematicError,
    DiematicConnectionError,
    DiematicParseError,
    DiematicResponseError,
)
import json

_LOGGER = logging.getLogger(__name__)


class DiematicBoilerClient:
    """Main class for handling connections with Diematic HTTP servers."""

    def __init__(
        self,
        host: str,
        port: int = 8080,
        base_path: str = '/diematic/',
        request_timeout: int = 8,
        session: aiohttp.client.ClientSession = None,
        tls: bool = False,
        username: str = None,
        password: str = None,
        verify_ssl: bool = False,
        user_agent: str = None,
    ) -> None:
        """Initialize connection with Diematic server."""
        self._session = session
        self._close_session = False

        self.base_path = base_path
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.request_timeout = request_timeout
        self.tls = tls
        self.verify_ssl = verify_ssl
        self.user_agent = user_agent

        self.diematic_uri = self._build_diematic_uri()

        self.session_sem = asyncio.Semaphore()

        if user_agent is None:
            self.user_agent = f'PythonDiematicClient{__version__}'

    def _build_diematic_uri(self):
        """Build a valid URI to connect to service"""
        scheme = 'https' if self.tls else 'http'

        return URL.build(
            scheme=scheme, host=self.host, port=self.port, path=self.base_path
        ).human_repr()

    def _message(self, operation: DiematicOperation, msg: dict) -> dict:
        """Build a request message to be sent to the server."""
        base = {'operation': operation}

        if not isinstance(msg, dict):
            msg = {}
        if 'data' not in list(msg):
            msg['data'] = None
        if operation == DiematicOperation.GET_CONFIG:
            msg['uri'] = 'config'
            msg['method'] = 'GET'
        elif operation == DiematicOperation.GET_VALUES:
            msg['uri'] = 'json'
            msg['method'] = 'GET'
        elif operation == DiematicOperation.SET_VALUE:
            parameter = msg['parameter']
            msg['uri'] = f'parameters/{parameter}'
            msg['method'] = 'POST'
            msg['data'] = json.dumps({'value': msg['value']})
        elif operation == DiematicOperation.GET_VALUE:
            parameter = msg['parameter']
            msg['uri'] = f'parameters/{parameter}'
            msg['method'] = 'GET'

        return always_merger.merge(base, msg)

    async def _request(
        self,
        data: dict,
        params: Optional[dict] = None,
    ) -> dict or list or str:
        """Handle a request to a Diematic server."""
        scheme = 'https' if self.tls else 'http'

        method = data['method']
        url = URL.build(
            scheme=scheme, host=self.host, port=self.port, path=self.base_path
        ).join(URL(data['uri']))

        auth = None
        if self.username and self.password:
            auth = aiohttp.BasicAuth(self.username, self.password)

        headers = {
            'User-Agent': self.user_agent,
            'Accept': 'application/json, text/plain, */*',
            'Connection': 'keep-alive',
            'Keep-Alive': 'timeout=5, max=100',
            'Cache-Control': 'max-age=0',
        }

        await self.session_sem.acquire()
        try:
            if self._session is None:
                self._session = aiohttp.ClientSession()
                self._close_session = True

            try:
                async with async_timeout.timeout(self.request_timeout):
                    response = await self._session.request(
                        method,
                        url,
                        auth=auth,
                        data=data['data'],
                        params=params,
                        headers=headers,
                        ssl=self.verify_ssl,
                    )
                    try:
                        if (response.status // 100) in [4, 5]:
                            content = response.read()

                            raise DiematicResponseError(
                                f'HTTP {response.status}',
                                {
                                    'content-type': response.headers.get('Content-type'),
                                    'message': content.decode('utf-8'),
                                    'status-code': response.status,
                                },
                            )

                        if 'content-type' in response.headers.keys():
                            type = response.headers['content-type']
                            if type.startswith('application/json'):
                                if 'content-length' in response.headers.keys():
                                    length = response.headers['content-length']
                                    content = await response.content.readexactly(int(length))
                                    return json.loads(content)
                                return await response.json()

                        return await response.text()

                    finally:
                        response.close()

            except asyncio.TimeoutError as exc:
                raise DiematicConnectionError(
                    'Timeout occurred while connecting to Diematic server.'
                ) from exc
            except (aiohttp.ClientError, SocketGIAError) as exc:
                raise DiematicConnectionError(
                    'Error occurred while communicating with Diematic server.'
                ) from exc
            finally:
                if self._close_session:
                    await self._session.close()
                    self._session = None
        finally:
            self.session_sem.release()

    async def execute(self, operation: DiematicOperation, message: dict) -> dict or list or str:
        """Send a request message to the server."""
        message = self._message(operation, message)
        return await self._request(message)

    async def boiler(self) -> Boiler:
        """Get boiler information from server."""
        response_data = await self.execute(DiematicOperation.GET_VALUES, {})

        try:
            boiler = Boiler.from_dict(response_data)
        except Exception as exc:
            raise DiematicParseError from exc

        return boiler

    async def config(self) -> list:
        """Bet Boiler registers configuration from server."""
        response_data = await self.execute(DiematicOperation.GET_CONFIG, {})

        return response_data

    async def update_boiler_register(self, parameter: str, value: int or float or str) -> DiematicStatus:
        """Update a register in the boiler."""
        try:
            await self.execute(DiematicOperation.SET_VALUE, {'parameter': parameter,'value': value})
            return DiematicStatus.OK
        except Exception as error:
            raise DiematicError(
                f'Cannot set register \'{parameter}\' to value {value}'
            ) from error

    async def read_boiler_register(self, parameter: str) -> dict:
        """Read single register from boiler and returns dict representation of returned json"""
        try:
            response_data = await self.execute(DiematicOperation.GET_VALUE, {'parameter': parameter})
            return response_data
        except Exception as error:
            raise DiematicError(
                f'Cannot read register \'{parameter}\''
            ) from error



"""Exceptions for Diematic."""


class DiematicError(Exception):
    """Generic Diematic exception."""


class DiematicConnectionError(DiematicError):
    """Diematic connection exception."""


class DiematicParseError(DiematicError):
    """Diematic parse exception."""


class DiematicResponseError(DiematicError):
    """Diematic response exception."""



# coding=UTF-8
"""Setup for the pipy package"""
import setuptools

with open('README.md', 'r', encoding='utf-8') as long_description_f:
	long_description = long_description_f.read()

setuptools.setup(
  name = 'diematic_client',
  version = '1.7',
  description = 'Asynchronous Python client for diematic-server HTTP server',
	long_description = long_description,
	long_description_content_type = 'text/markdown; charset=UTF-8',
  author = 'Ignacio Hernández-Ros',
  author_email = 'ignacio@hernandez-ros.com',
  packages = ['diematic_client'],
  license='LGPL',
  url = 'https://github.com/IgnacioHR/diematic_client',
  download_url = 'https://github.com/IgnacioHR/diematic_client/archive/refs/tags/v1.7.tar.gz',
  keywords = ['python', 'home-automation', 'iot', 'restful', 'de-dietrich', 'diematic', 'home-assistant'],
  install_requires=[
		'aiohttp==3.8.1',
		'deepmerge==1.0.1',
  ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
		'Environment :: Plugins',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)',
		'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.7',
		'Topic :: Home Automation',
	],
)